{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : Discriminative sorting
    Copyright   : Fritz Henglein, 2007
    License     : All Rights Reserved

    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module DSort where

import Order 
import Array (accumArray, assocs)
import List (groupBy)

-- Generic definition of stable sorting function induced by ordering
-- Note clauses for SumL, PairL, Map0: Motivate notion of discriminator
dsort :: Order k -> [(k, v)] -> [(k, v)]

dsort r []               = []
dsort r xs @ [(k, v)]    = xs
dsort (Nat n) xs         = bsortNat n xs
dsort Triv xs            = xs
dsort (SumL r1 r2) xs     =  
 [ x | (_, x) <- dsort r1 [ (k, x) | x @ (Left k, _) <- xs ] ]  ++ 
 [ y | (_, y) <- dsort r2 [ (k', y) | y @ (Right k', _) <- xs ] ]
dsort (PairL r1 r2) xs    =  
 [ x | (_, x) <- dsort r1 [ (k1, x) | x @ ((k1, k2), v) <- xs' ] ] 
 where 
   xs' = [ x | (_, x) <- dsort r2 [ (k2, x) | x @ ((k1, k2), v) <- xs ] ]
dsort (MapO f r) xs       = 
 [ x | (_, x) <- dsort r [ (f k, x) | x @ (k, v) <- xs ] ]
dsort (Inv r) xs         = reverse (dsort r (reverse xs))

bsortNat n xs = [ (k, v) | (k, vs) <- blocks, v <- reverse vs ] 
     where blocks = assocs (accumArray (\vs v -> v : vs) [] (0, n-1) xs)

-- Sorting on keys only
ksort :: Order t -> [t] -> [t]
ksort r xs = [ k | (k, _) <- dsort r [ (k, ()) | k <- xs ] ] 
