{-# OPTIONS -fglasgow-exts #-}

----------------------------------------------------------------------------------------
{-| Module      : Disc: Equivalence discriminators
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved
    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
----------------------------------------------------------------------------------------
module Disc where

import Order
import Equiv
import NatDisc
import SDisc (split, bucket)

-- Generic equivalence discriminator
disc :: Equiv k -> [(k, v)] -> [[v]]
disc e = disc' (normalize e)

disc' :: Equiv k -> [(k, v)] -> [[v]]
disc' e []               = []
disc' e [(k, v)]         = [[v]]
disc' (Equiv (Nat n)) xs = if n > 65535 then disc eqInt32 xs else discNat16 xs
disc' (Equiv Triv) xs    = [map snd xs]
disc' (SumE e1 e2) xs    =  disc' e1 lefts ++ disc' e2 rights 
          where (lefts, rights) = split xs
disc' (PairE e1 e2) xs   = [ vs | ys <-disc' e1 [ (k1, (k2, v)) | ((k1, k2), v) <- xs ],
                                  vs <- disc' e2 ys ]
disc' (MapE f e) xs      = disc' e [ (f k, v) | (k, v) <- xs ] 
disc' (BagE e) xs        = discColl updateBag e xs
         where updateBag vs v = v : vs
disc' (SetE e) xs         = discColl updateSet e xs
         where updateSet [] w         = [w]
               updateSet vs@(v : _) w = if v == w then vs else w : vs
discColl update e xss = disc' (list (Equiv (Nat (length memGroups)))) yss
         where
         (kss, vs)   = unzip xss                
         memGroups   = disc' e (groupNum kss)   
         sigs        = bucket (length kss) update (groupNum memGroups)
         yss         = zip sigs vs      

-- Group number representation of list of groups
groupNum :: [[k]] -> [(k, Int)]
groupNum xss = [ p | (xs, i) <- zip xss [0..], p <- zip xs (repeat i) ]

-- Partitioning from discrimination
part :: Equiv t -> [t] -> [[t]]
part e xs = disc e [ (x, x) | x <- xs ]

-- Representatives from partitioning
reps :: Equiv t -> [t] -> [t]
reps e xs = [ head ys | ys <- part e xs ]
