{-# OPTIONS -fglasgow-exts #-}
--------------------------------------------------------------------------------
{-| Module      : Order discriminator combinators
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved
    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
--------------------------------------------------------------------------------
module DiscComb where

import SDisc (sdiscNat)

type SDisc k v = [(k, v)] -> [[v]]

-- Primitive discriminators
-- sdiscNat :: Int -> SDisc Int v from Disc

sdiscTriv :: SDisc k v
sdiscTriv xs = [[ v | (_, v) <- xs ]]

-- Discriminator combinators
sdiscSumL :: SDisc k1 v -> SDisc k2 v -> SDisc (Either k1 k2) v
sdiscSumL d1 d2 xs  =  
    d1 [ (k1, v1) | (Left k1, v1) <- xs ] 
 ++ d2 [ (k2, v2) | (Right k2, v2) <- xs ]

sdiscPairL :: SDisc k1 (k2, v) -> SDisc k2 v -> SDisc (k1, k2) v
sdiscPairL d1 d2 xs = 
 [ vs | ys <- d1 [ (k1, (k2, v)) | ((k1, k2), v) <- xs ],
        vs <- d2 ys ]

sdiscMapO :: (k1 -> k2) -> SDisc k2 v -> SDisc k1 v
sdiscMapO f d xs  = d [ (f k, v) | (k, v) <- xs ] 

sdiscInv :: SDisc k v -> SDisc k v
sdiscInv d xs       = reverse (d xs)

-- Discriminator-parameterized partitioning and sorting functions
spartD :: SDisc t t -> [t] -> [[t]]
spartD d xs = d [ (x, x) | x <- xs ]

dsortD :: SDisc t t -> [t] -> [t]
dsortD d xs = [ y | ys <- spartD d xs, y <- ys ]

usortD :: SDisc t t -> [t] -> [t]
usortD d xs = [ head ys | ys <- spartD d xs ]

-- Test of order discriminator combinators
e = sdiscPairL (sdiscNat 100) (sdiscTriv) [((5, "AB"), "XXX"), ((8, "CD"), "YYY"), ((5, "EF"), "ZZZ")]