{-# OPTIONS -fglasgow-exts #-}
--------------------------------------------------------------------------------
{-| Module      : Discriminators for number types
    Copyright   : Fritz Henglein, 2009
    License     : All Rights Reserved
    Maintainer  : Fritz Henglein
    Stability   : Unknown
    Portability : Unknown
-}
--------------------------------------------------------------------------------
module NatDisc where

import Array (accumArray, elems)
import Control.Monad
import Data.Array.IO
import System.IO.Unsafe
import Data.IORef

-- discNat size : Returns basic multiset discriminator for equality on [0..size] 
-- based on preallocated and initialized bucket table T[0..size].
-- ! The discriminator is occurrence-order preserving and reverse stable !
discNat :: Int -> [(Int, v)] -> [[v]]
discNat size =
  unsafePerformIO (
  do { table <- newArray (0, size) [] :: IO (IOArray Int [v]) ;
       let discNat xs = unsafePerformIO (
            do { ks <- foldM (\keys (k, v) -> 
                               do { vs <- readArray table k ;
                                    case vs of { [] -> do { writeArray table k [v] ;
                                                            return (k : keys) } ;
                                                 _ ->  do { writeArray table k (v : vs) ;
                                                            return keys } } } )
                            [] xs ;
                 foldM (\vss k -> do { elems <- readArray table k ;
                                       writeArray table k [] ;
                                       return (elems : vss) }) 
                       [] ks } )
       in return discNat } )  

-- Basic multiset discriminator with table of size 2^16 = 65536.  
-- Keys must have type Int and be in range [0 .. 2^16-1]
{-# NOINLINE discNat16 #-} 
discNat16 :: [(Int, v)] -> [[v]]
discNat16 = discNat 65535

